import json
import logging
import time

import boto3
import botocore
import click
from boto.awslambda.exceptions import ResourceNotFoundException
from botocore.exceptions import ClientError, NoCredentialsError
from typing import List, Any, Tuple, Optional

from inoft_vocal_engine.cloud_providers.aws.core_clients import AwsCoreClients
from inoft_vocal_engine.safe_dict import SafeDict


class AwsCore(AwsCoreClients):
    def __init__(self, clients_to_load: List[int]):
        super().__init__(clients_to_load=clients_to_load)

        self.tags = {"framework": "InoftVocalFramework"}

        self._iam_role = None
        self._credentials_arn = None
        self._aws_account_id = None

        self._session_unix_time = None
        self._session_count_created_statement_ids = 0

        self.role_name = "InoftVocalFrameworkLambdaExecution"
        self.http_methods = ['ANY']

        self.default_description = "Created with the Inoft Vocal Framework"

    def create_s3_bucket_if_missing(self, bucket_name: str, region_name: str) -> bool:
        try:
            self.s3_client.head_bucket(Bucket=bucket_name)
            return True
        except ClientError as e:
            click.echo(f"Trying to create a new S3 Bucket with name {click.style(text=bucket_name, fg='yellow', bold=True)}"
                       f" in region {click.style(text=region_name, fg='yellow', bold=True)}")
            available_regions_for_s3 = self.boto_session.get_available_regions(service_name="s3")
            if region_name not in available_regions_for_s3:
                raise Exception(f"The region {region_name} was not available for s3. Here is the available regions : {available_regions_for_s3}")

            self.s3_client.create_bucket(Bucket=bucket_name, CreateBucketConfiguration={"LocationConstraint": region_name})
            click.echo(f"Completed creation of the new bucket.")
            return False

    def _s3_prepare_upload_operation(self, bucket_name: str, region_name: str, access_should_be_public: bool = True) -> Tuple[bool, dict]:
        # If an error happen while uploading to S3, then the upload will not be
        # successful. We use that as our way to send a success response.
        extra_args = dict()
        if access_should_be_public is True:
            extra_args["ACL"] = "public-read"

        try:
            self.create_s3_bucket_if_missing(bucket_name=bucket_name, region_name=region_name)
            return True, extra_args
        except NoCredentialsError as e:
            click.echo(f"Almost there ! You just need to configure your AWS credentials."
                       f"\nYou can follow the official documentation (you will need to install the awscli by running pip install awscli) "
                       f"then go to https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html#configuration")
        except Exception as e:
            click.echo(f"Error while trying to create the S3 bucket before an upload operation : {e}")
        return False, extra_args

    def put_object_to_s3(self, object_item: Any, object_key_name: str, bucket_name: str, region_name: str, access_should_be_public: bool = True) -> bool:
        preparation_success, extra_kwargs = self._s3_prepare_upload_operation(
            bucket_name=bucket_name, region_name=region_name, access_should_be_public=access_should_be_public
        )
        if preparation_success is True:
            try:
                self.s3_client.put_object(Body=object_item, Bucket=bucket_name, Key=object_key_name, **extra_kwargs)
                return True
            except Exception as e:
                click.echo(f"Error while putting an object to the S3 bucket : {e}")
        return False

    def upload_file_to_s3(self, filepath: str, object_key_name: str, bucket_name: str, region_name: str, access_should_be_public: bool = True) -> bool:
        preparation_success, extra_args = self._s3_prepare_upload_operation(
            bucket_name=bucket_name, region_name=region_name, access_should_be_public=access_should_be_public
        )
        if preparation_success is True:
            try:
                self.s3_client.upload_file(Filename=filepath, Bucket=bucket_name, Key=object_key_name, ExtraArgs=extra_args)
                return True
            except Exception as e:
                click.echo(f"Error while uploading file to the S3 bucket : {e}")
        return False

    def remove_from_s3(self, file_name: str, bucket_name: str) -> bool:
        try:
            self.s3_client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                return False

        try:
            self.s3_client.delete_object(Bucket=bucket_name, Key=file_name)
            return True
        except (botocore.exceptions.ParamValidationError, botocore.exceptions.ClientError):
            return False

    def get_all_files_from_s3_folder(self, folderpath: str, bucket_name: str):
        response = self.s3_client.list_objects_v2(
            Bucket=bucket_name,
            Prefix=folderpath,
            MaxKeys=1000,
        )
        return response

    def get_lambda_function_arn(self, function_name_or_arn: str) -> Optional[str]:
        try:
            return SafeDict(self.lambda_client.get_function(FunctionName=function_name_or_arn)).get("Configuration").get("FunctionArn").to_str(default=None)
        except Exception as e:
            return None

    def lambda_function_exist(self, function_name_or_arn: str) -> bool:
        return True if self.get_lambda_function_arn(function_name_or_arn=function_name_or_arn) is not None else False

    def rest_api_gateway_exist(self, api_id: str) -> bool:
        try:
            response = self.api_gateway_client.get_rest_api(restApiId=api_id)
            return True
        except Exception as e:
            print(f"Rest API Gateway not found."
                  f"  --api_id:{api_id}"
                  f"  --error:{e}")
            return False

    def api_gateway_v2_url(self, api_id: str) -> str or None:
        try:
            return SafeDict(self.api_gateway_v2_client.get_api(ApiId=api_id)).get("ApiEndpoint").to_str(default=None)
        except Exception as e:
            return None

    def get_lambda_function_versions(self, function_name: str) -> list:
        try:
            response = self.lambda_client.list_versions_by_function(FunctionName=function_name)
            return response.get('Versions', list())
        except Exception as e:
            click.echo(f"Lambda function {function_name} not found. Error : {e}")
            return list()

    def create_lambda_function(self, bucket=None, s3_key: str = None, local_zip: bytes = None,
                               function_name: str = None, handler=None,
                               description: str = "Inoft Vocal Engine Deployment",
                               timeout: int = 30, memory_size: int = 512, publish: bool = True, runtime=None):
        """
        Given a bucket and key (or a local path) of a valid Lambda-zip, a function name and a handler, register that Lambda function.
        """
        kwargs = dict(
            FunctionName=function_name,
            Runtime=runtime,
            Role=self.credentials_arn,
            # "arn:aws:iam::631258222318:role/service-role/alexa_hackaton-cite-des-sciences_fakenews-challeng-role-ck3sxsyt",  # self.credentials_arn,
            # todo: make dynamic and create automaticly role if missing
            Handler=handler,
            Description=description,
            Timeout=timeout,
            MemorySize=memory_size,
            Publish=publish,
            Layers=[
                "arn:aws:lambda:eu-west-3:631258222318:layer:inoft-vocal-framework_v0-8dot3:2",
                # "arn:aws:lambda:eu-west-3:631258222318:layer:ffmpeg:2",
            ]
            # Those ARNs point to the required lambda layers to use the framework. Please do not modify them.
        )
        if local_zip is not None:
            kwargs['Code'] = {
                'ZipFile': local_zip
            }
        else:
            kwargs['Code'] = {
                'S3Bucket': bucket,
                'S3Key': s3_key
            }

        response = self.lambda_client.create_function(**kwargs)

        lambda_arn = response["FunctionArn"]
        version = response['Version']

        click.echo(f"Created new lambda function {function_name} with arn of {lambda_arn}")
        return lambda_arn

    def update_lambda_function_code(self, lambda_arn: str, object_key_name: str, bucket_name: str) -> dict:
        response = self.lambda_client.update_function_code(
            FunctionName=lambda_arn,
            S3Bucket=bucket_name,
            S3Key=object_key_name,
            Publish=True,
        )
        return response

    def update_lambda_function_configuration(self, function_name: str, handler_function_path: str = None,
                                             lambda_layers_arns: list = None) -> dict:
        kwargs = dict(FunctionName=function_name)
        if handler_function_path is not None:
            kwargs["Handler"] = handler_function_path
        if lambda_layers_arns is not None and len(lambda_layers_arns) > 0:
            kwargs["Layers"] = lambda_layers_arns

        response = self.lambda_client.update_function_configuration(**kwargs)
        return response

    def create_api_gateway(self, lambda_arn: str, api_gateway_name: str, route_names: List[str], description: str = None) -> Optional[str]:
        # todo: fix bug if id of api gateway is present in file, but the api has been deleted, it will not try to recreate it
        api_creation_response = SafeDict(self.api_gateway_client.create_rest_api(
            name=api_gateway_name, description=description or self.default_description
        ))
        api_id = api_creation_response.get("id").to_str(default=None)

        if api_id is not None:
            self.create_and_setup_api_methods(api_id=api_id, lambda_arn=lambda_arn, route_names=route_names)
            return api_id
        return None

    def get_resource_id_by_path(self, api_id: str, path: str) -> str:
        response = SafeDict(self.api_gateway_client.get_resources(restApiId=api_id))
        response_items = response.get("items").to_list()
        for resource in response_items:
            resource_safedict = SafeDict(resource)
            resource_path = resource_safedict.get("path").to_str(default=None)
            if resource_path is not None and resource_path == path:
                resource_id = resource_safedict.get("id").to_str(default=None)
                if resource_id is not None:
                    return resource_id
        return None

    def get_first_resource_id(self, api_id) -> str or None:
        response = SafeDict(self.api_gateway_client.get_resources(restApiId=api_id))
        response_items = response.get("items").to_list()
        for item in response_items:
            item_id = SafeDict(item).get("id").to_str(default=None)
            if item_id is not None:
                return item_id
        return None

    def _create_resource(self, api_id: str, path: str, parent_resource_id: str) -> str:
        resource_creation_response = SafeDict(self.api_gateway_client.create_resource(
            restApiId=api_id,
            parentId=parent_resource_id,
            pathPart=path
        ))
        created_resource_id = resource_creation_response.get('id').to_str(default=None)
        if created_resource_id is not None:
            # todo: when creating the api, add an AWS_PROXY (Lambda Proxy) integration
            pass
            """integration_creation_response = SafeDict(self.api_gateway_client.put_integration(
                restApiId=api_id,
                resourceId=created_resource_id,
                httpMethod='ANY',
                type='AWS_PROXY'
            ))"""
        return created_resource_id

    def put_method(api_id: str, parent_id: str, method: str, f: str, path: str = ""):
        apigateway = boto3.client('apigateway')

        apigateway.put_method(restApiId=api_id, resourceId=parent_id, httpMethod=method, authorizationType='NONE')

        apigateway.put_method_response(
            restApiId=api_id, resourceId=parent_id, httpMethod=method,
            statusCode="200", responseModels={"application/json": "Empty"})

        region = boto3.session.Session().region_name
        acct = boto3.client('sts').get_caller_identity().get('Account')
        uri = f"arn:aws:apigateway:{region}:lambda:path/2015-03-31/functions/arn:aws:" \
              f"lambda:{region}:{acct}:function:{f}/invocations"

        apigateway.put_integration(
            restApiId=api_id, resourceId=parent_id, httpMethod=method,
            type="AWS", integrationHttpMethod="POST", uri=uri
        )

        apigateway.put_integration_response(
            restApiId=api_id, resourceId=parent_id, httpMethod=method, statusCode="200",
            responseTemplates={'application/json': ""})

        source_arn = f"arn:aws:execute-api:{region}:{acct}:{api_id}/*/{method}/{path}"

        lambda_api = boto3.client('lambda')
        lambda_api.add_permission(
            FunctionName=f"arn:aws:lambda:{region}:{acct}:function:{f}",
            StatementId=uuid.uuid4().hex,
            Action="lambda:InvokeFunction",
            Principal="apigateway.amazonaws.com",
            SourceArn=source_arn
        )

    def _setup_method(self, api_id: str, created_resource_id: str, http_method: str, lambda_arn: str):
        # An authorizationType of NONE means that the method will have open access
        method_creation_response = SafeDict(self.api_gateway_client.put_method(
            restApiId=api_id, resourceId=created_resource_id, httpMethod=http_method, authorizationType="NONE")
        )

        region = self.boto_session.region_name
        account = boto3.client('sts').get_caller_identity().get('Account')
        uri = f"arn:aws:apigateway:{region}:lambda:path/2015-03-31/functions/arn:aws:" \
              f"lambda:{region}:{account}:function:{lambda_arn}/invocations"

        self.api_gateway_client.put_integration(
            restApiId=api_id, resourceId=created_resource_id, httpMethod='ANY',
            type='AWS', integrationHttpMethod='POST', uri=uri
        )

    def create_and_setup_api_methods(self, api_id: str, lambda_arn: str, route_names: List[str]):
        root_resource_id = self.get_resource_id_by_path(api_id=api_id, path='/')
        if root_resource_id is None:
            raise Exception(f"No route named / has been found in the api with id {api_id}")

        self._setup_method(api_id=api_id, created_resource_id=root_resource_id,
                           http_method="ANY", lambda_arn=lambda_arn)

        for route_name in route_names:
            self.boto_session.get_credentials()
            created_resource_id = self._create_resource(api_id=api_id, path=route_name, parent_resource_id=root_resource_id)
            if created_resource_id is not None:
                self._setup_method(api_id=api_id, created_resource_id=created_resource_id,
                                   http_method="ANY", lambda_arn=lambda_arn)

                """
                apigateway.put_integration(
                    restApiId=api_id, resourceId=parent_id, httpMethod=method,
                    type="AWS", integrationHttpMethod="POST", uri=uri
                )
    
                response = self.api_gateway_client.create_route(
                    ApiId=api_id,
                    AuthorizationType="None",
                    RouteKey=f"ANY /{route_name}",
                    # A space is required between the HTTP method and the /
                    Target=f"integrations/{integration_id}",
                )
                self.add_lambda_permission_to_call_api_resource(lambda_arn=lambda_arn, api_id=api_id, route_key=route_name)
                api_route_url = f"{api_gateway_root_url}/{route_name}"
                click.echo(f"Api route {click.style(route_name, fg='green')} creation complete accessible on {api_route_url}")
                # self.settings.get_set("deployment", {}).get_set("endpoints", {}).put(
                #    route_names_to_settings_keys[route_name], api_route_url).reset_navigated_dict()
                # We reset the navigated dict after a final put
                """

    def add_lambda_permission_to_call_api_resource(self, lambda_arn: str, api_id: str, route_key: str):
        response = self.lambda_client.add_permission(
            FunctionName=lambda_arn,
            StatementId=self.get_session_new_statement_id(),
            Action="lambda:InvokeFunction",
            Principal="apigateway.amazonaws.com",
            SourceArn=f"arn:aws:execute-api:{self.boto_session.region_name}:{self.aws_account_id}:{api_id}/*/*/{route_key}",
        )

    @property
    def iam_role(self) -> SafeDict:
        if self._iam_role is None:
            try:
                self._iam_role = SafeDict(self.iam_client.get_role(RoleName=self.role_name)["Role"])
            except botocore.exceptions.ClientError:
                # If a ClientError happened while getting the role, it means he do not exist.
                logging.debug(f"Creating {self.role_name} IAM Role..")

                attach_policy_obj = json.loads(self.attach_policy)
                assume_policy_obj = json.loads(self.assume_policy)

                self._iam_role = SafeDict(self.iam_client.create_role(
                    RoleName=self.role_name, AssumeRolePolicyDocument=self.assume_policy)["Role"])

                # create or update the role's policies if needed
                policy = self.iam_resource.RolePolicy(self.role_name, "inoft-vocal-permissions")
                try:
                    if policy.policy_document != attach_policy_obj:
                        logging.debug(f"Updating inoft-vocal-permissions policy on {self.role_name} IAM Role.")
                        policy.put(PolicyDocument=self.attach_policy)

                except botocore.exceptions.ClientError:
                    logging.debug(f"Creating inoft-vocal-permissions policy on {self.role_name} IAM Role.")
                    policy.put(PolicyDocument=self.attach_policy)

                role_policy_dict = self._iam_role.get("AssumeRolePolicyDocument").to_dict()
                if role_policy_dict != assume_policy_obj:
                    if (SafeDict(role_policy_dict["Statement"][0]).get("Principal").get("Service").to_any()
                            != assume_policy_obj["Statement"][0]["Principal"]["Service"]):
                        logging.debug(f"Updating assume role policy on {self.role_name} IAM Role.")
                        self.iam_client.update_assume_role_policy(RoleName=self.role_name,
                                                                  PolicyDocument=self.assume_policy)

        return self._iam_role

    @property
    def credentials_arn(self):
        if self._credentials_arn is None:
            self._credentials_arn = self.iam_role.get("Arn").to_str()
        return self._credentials_arn

    @credentials_arn.setter
    def credentials_arn(self, credentials_arn) -> None:
        self._credentials_arn = credentials_arn

    @property
    def aws_account_id(self) -> str:
        if self._aws_account_id is None:
            self._aws_account_id = boto3.client("sts").get_caller_identity().get("Account")
        return self._aws_account_id

    def get_session_new_statement_id(self) -> str:
        # No restriction is set on the statement id. To created my type of ids, i take the same unix
        # time across the session, and each time we generated a new statement id, i add 1 to the unix time.
        # Dumb and simple, but it assure that i will never have the same id twice.
        if self._session_unix_time is None:
            self._session_unix_time = time.time()
        self._session_count_created_statement_ids += 1

        # After turning the modified unix time into a string, we remove any point, that would have
        # been used to  separate the decimals, because AWS cannot accept a point in a id string.
        return str(self._session_unix_time + self._session_count_created_statement_ids).replace(".", "")

if __name__ == "__main__":
    Core().upload_file_to_s3(
        "F:\Inoft\hackaton cite des sciences 1\lambda_project\inoft_vocal_engine\cli\core\core_clients.py",
        "letestduframeworkinoft", "eu-west-3")




from typing import Optional, Any
from pydantic import BaseModel


class BaseAction(BaseModel):
    _TYPE_NAME = "base"

    isNew: bool = False
    nodeId: str
    actionType: str
    _FIELDS_TO_EXCLUDE: set = {"isNew", "nodeId"}

    def to_database(self) -> dict:
        return self.dict(exclude=self._FIELDS_TO_EXCLUDE)


class SayTextAction(BaseAction):
    _TYPE_NAME = "sayText"

    textContentItemId: Optional[str]
    loadedTextToSay: Optional[str]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        super()._FIELDS_TO_EXCLUDE.add("loadedTextToSay")

    # todo: deprecated
    @classmethod
    def make_new(cls, text_content_id: Optional[str], **kwargs):
        return cls(textContentItemId=text_content_id, **kwargs)

class UserDataMemorizeVariableAction(BaseAction):
    _TYPE_NAME = "userDataMemorizeVariable"

    variableName: Optional[str]
    variableType: Optional[str]


actions_keys_switch = {
    BaseAction._TYPE_NAME: BaseAction,
    SayTextAction._TYPE_NAME: SayTextAction,
    UserDataMemorizeVariableAction._TYPE_NAME: UserDataMemorizeVariableAction
}

def get_appropriate_action_from_dict(action_data_dict: dict) -> Optional[BaseAction]:
    if "actionType" in action_data_dict.keys():
        action_class_type = actions_keys_switch.get(action_data_dict["actionType"], None)
        if action_class_type is not None:
            return action_class_type(**action_data_dict)
    return None






from socket import *

def start_socket():
    while True:
        print("hey")
        PySocket = socket(AF_INET, SOCK_DGRAM)
        PySocket.bind(('127.0.0.1', 9999))
        PySocket.send(b"Salut ! :)")
        data, client = PySocket.recvfrom(1024)
        print(data)


# todo: make the socket reading async
if __name__ == "__main__":
    start_socket()




import os
from inoft_vocal_engine import InoftSkill, InoftCondition, InoftRequestHandler, InoftStateHandler, InoftDefaultFallback


class YesCondition(InoftCondition):
    def can_handle(self) -> bool:
        return self.is_in_intent_names(["AMAZON.YesIntent", "YesIntent"])


class StartRequestHandler(InoftRequestHandler):
    def can_handle(self) -> bool:
        return self.is_launch_request()

    def handle(self) -> dict:
        self.say("Hey ! Are you ready for your first fact about the Inoft Vocal Framework ?")
        self.memorize_session_then_state(SayFactOnYesStateHandler)
        return self.to_platform_dict()


class GetFactRequestHandler(InoftRequestHandler):
    def can_handle(self) -> bool:
        return self.is_in_intent_names("GetFact")

    def handle(self) -> dict:
        from messages import MSGS_START_FACT, MSGS_END_FACT, MSGS_FACTS, MSGS_NEW_FACT
        self.say(f"{MSGS_START_FACT.pick()} {MSGS_FACTS.pick()} {MSGS_END_FACT.pick()} {MSGS_NEW_FACT.pick()}")
        self.memorize_session_then_state(SayFactOnYesStateHandler)
        return self.to_platform_dict()


class SayFactOnYesStateHandler(InoftStateHandler):
    def handle(self) -> dict:
        if YesCondition(self).can_handle():
            return GetFactRequestHandler(self).handle()
        else:
            self.say("I'm supposed to be here to give you facts. Yet if i want i can give you some infos about this application. What do you say ? Yes ?")
            return self.to_platform_dict()


class GetFactOrCloseStateHandler(InoftStateHandler):
    def handle(self) -> dict:
        if YesCondition(self).can_handle():
            return GetFactRequestHandler(self).handle()
        else:
            self.say("Oki ! Have a great day !")
            self.end_session()
            return self.to_platform_dict()


class WantSomeInformationsStateHandler(InoftStateHandler):
    def handle(self):
        if YesCondition(self).can_handle():
            self.say(f"The application to thank for us talking to each other, has been developed in Python with the Inoft Vocal Framework, "
                     f"and work both for me {'Alexa, and also Google Assistant' if self.is_alexa else 'Google Assistant, and also Alexa'}")
        else:
            self.say("Then if you do not want a fact, i'm going to close myself. So, do you want a fact ?")
        return self.to_platform_dict()


class DefaultFallback(InoftDefaultFallback):
    def handle(self):
        self.say("I did not followed you on that one. Would like me to tell you a fact about the Inoft Vocal Framework ?")
        self.memorize_session_then_state(SayFactOnYesStateHandler)
        return self.to_platform_dict()


def lambda_handler(event, context):
    skill_builder = InoftSkill(settings_yaml_filepath=os.path.join(os.path.dirname(os.path.abspath(__file__)), "app_settings.yaml"))
    skill_builder.add_request_handler(StartRequestHandler)
    skill_builder.add_request_handler(GetFactRequestHandler)
    skill_builder.add_state_handler(SayFactOnYesStateHandler)
    skill_builder.add_state_handler(GetFactOrCloseStateHandler)
    skill_builder.add_state_handler(WantSomeInformationsStateHandler)
    skill_builder.set_default_fallback_handler(DefaultFallback)
    return skill_builder.handle_any_platform(event=event, context=context)

if __name__ == "__main__":
    from inoft_vocal_engine.inoft_vocal_framework.platforms_handlers.simulator import Simulator
    event_, context_ = Simulator(event_type="alexa/start").get_event_and_context()
    print(f"\n\nFinal Output : {lambda_handler(event=event_, context=context_)}")



import os
import time
import zipfile
from pathlib import Path
from typing import Optional

import botocore
import click
from botocore.exceptions import ClientError
from click import ClickException

# todo: fix issue where when deploying we do not check if the api with the id found in the settings file do exit
from inoft_vocal_engine.cloud_providers.aws.aws_core import AwsCore
from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path


class DeployHandler(AwsCore):
    def __init__(self):
        click.echo("Initializing AWS clients... (this can take a few seconds)")
        start_time = time.time()
        super().__init__(clients_to_load=[AwsCore.CLIENT_S3, AwsCore.CLIENT_LAMBDA, AwsCore.CLIENT_API_GATEWAY,
                                          AwsCore.CLIENT_IAM, AwsCore.CLIENT_STS, AwsCore.CLIENT_BOTO_SESSION])
        click.echo(f"Took {click.style(text=f'{round(time.time() - start_time, 2)}s', fg='white')} to initiate AWS clients")
        # todo: make the initialization even more asynchronous, by calling the handle function while the AWS ressources are initializing.
        #  Right now, the initialization of the resources are async, but we must wait for their completion in order to call the handle function.

    def deploy(self, lambda_files_root_folderpath: str, stage_name: str, bucket_name: str, bucket_region_name: str,
               lambda_name: str, lambda_handler: str, runtime="python3.7", upload_zip: bool = True,
               existing_zip_filepath: Optional[str] = None, lambda_description: str = "Inoft Vocal Engine Deployment",
               lambda_timeout_seconds: int = 30, lambda_memory_size: int = 512, publish: bool = True):

        # Make sure this isn't already deployed.
        """deployed_versions = self.get_lambda_function_versions(lambda_name)
        if len(deployed_versions) > 0:
            click.echo(f"This lambda function for this application is {click.style('already deployed', fg='green')}" +
                  f" If some part of the full deployment failed or you changed some resource name, and you want to do it again, type Y."
                  f" Otherwise, use the {click.style('inoft update', bold=True)} command")
            if not click.confirm("Do you want to redo the deployment ?"):
                click.echo(f"Ok ! Remember to use the {click.style('inoft update', bold=True)} command !")
                exit()
        """

        zip_filepath, upload_success = None, None
        if upload_zip is True:
            if existing_zip_filepath is None:
                # Create the Lambda Zip
                zip_filepath = self.create_package(to_zip=True, to_s3=False,
                                                   app_folder_path=lambda_files_root_folderpath,
                                                   stage_name=stage_name, bucket_name=bucket_name, bucket_region_name=bucket_region_name)
            else:
                if not os.path.isfile(existing_zip_filepath):
                    raise Exception(f"Existing app project zip file do not exist at filepath : {existing_zip_filepath}")
                zip_filepath = existing_zip_filepath

            # Upload it to S3
            click.echo("Uploading the lambda zip file to S3")
            upload_success = self.upload_file_to_s3(filepath=zip_filepath, object_key_name=Path(zip_filepath).name,
                                               bucket_name=bucket_name, region_name=bucket_region_name)
            if not upload_success:
                raise ClickException("Unable to upload to S3. Look in the logs what caused the errors,"
                                     " and modify your app_settings file in order to fix the issue."
                                     "\nCommon issues include :"
                                     "\n - Not having setup your credentials correctly (run the command : aws configure)"
                                     "\n - Trying to use a S3 bucket name that is not available."
                                     "\n - Having modified the app_settings file, but not have saved it with CTRL+S")
            click.echo("Uploading completed.")

        src_folderpath = os.path.join(get_inoft_vocal_engine_root_path(), "web_interface/static/dist")
        self.upload_folder_to_s3(src_folderpath=src_folderpath, s3_target_folderpath="prod/static/dist",
                                 bucket_name=bucket_name, bucket_region_name=bucket_region_name)

        return True

        # Remove the project zip from S3.
        if not source_zip:
            self.remove_uploaded_zip()

    def deploy_from_s3_files(self, s3_bucket_name: str, s3_dirpath: str):
        self.s3_client
        try:
            lambda_arn = self.get_lambda_function_arn(function_name_or_arn=lambda_name)
            click.echo(f"Using the existing lambda function {lambda_name}")
            if upload_success is True:
                self.update_lambda_function_code(lambda_arn=lambda_arn, object_key_name=Path(zip_filepath).name, bucket_name=bucket_name)
                self.update_lambda_function_configuration(function_name=lambda_arn, handler_function_path=lambda_handler)
                click.echo(f"Updated the code of the lambda with arn {lambda_arn}")

        except botocore.exceptions.ClientError:
            try:
                lambda_arn = self.create_lambda_function(bucket=bucket_name, s3_key=Path(zip_filepath).name, function_name=lambda_name,
                                                         handler=lambda_handler, description=lambda_description,
                                                         timeout=lambda_timeout_seconds, memory_size=lambda_memory_size,
                                                         runtime=runtime, publish=publish)
                click.echo(f"Created the lambda function : {lambda_name}")
            except Exception as e:
                click.echo(f"Error while updating the function by giving it a S3 path. Trying to update the"
                           f"function with a byte_stream of the zip file of the deployment package : {e}")

                with open(zip_filepath, mode="rb") as file_stream:
                    byte_stream = file_stream.read()

                lambda_arn = self.create_lambda_function(local_zip=byte_stream, function_name=lambda_name,
                                                         handler=lambda_handler, description=lambda_description,
                                                         timeout=lambda_timeout_seconds, memory_size=lambda_memory_size,
                                                         runtime=runtime, publish=publish)
                click.echo(f"Created the lambda function : {lambda_name}")

        # Create and configure the API Gateway
        need_to_create_api = False
        # api_gateway_id = self.settings.settings.get("deployment").get("apiGatewayId").to_str(default=None)
        api_gateway_id = "1g9v1ilt6b"  # None  # todo: store the id of the api in a database
        if api_gateway_id is None or self.rest_api_gateway_exist(api_gateway_id) is None:
            api_id = self.create_api_gateway(lambda_arn=lambda_arn, lambda_name=lambda_name, route_names=['{proxy+}'])
            click.echo(f"Created a new {click.style(text='API Gateway', bold=True)}")
            print(f"api_id = {api_id}")
            # self.settings.settings.get("deployment").put("apiGatewayId", api_id)
            # self.settings.save_settings()
        else:
            click.echo(f"Using the existing ApiGatewayV2 with id {click.style(text=api_gateway_id, bold=True, fg='green')}")


    def upload_folder_to_s3(self, src_folderpath: str, s3_target_folderpath: str, bucket_name: str, bucket_region_name: str):
        for root_dirpath, dirs, filenames in os.walk(src_folderpath, topdown=True):
            relative_root_dirpath = root_dirpath.replace(src_folderpath, "").replace("\\", "/").lstrip("/")
            # We need to remove potential leading backslashes and slashes, otherwise, when we will use the os.path.join
            # function, it will remove the first element of the path we are trying to combine the relative_root_dirpath.
            # For example, if we did os.path.join("prod", "\\folder\\file.py") we would end up with \\folder\\file.py
            # where as doing os.path.join("prod", "folder\\file.py") will give us prod\\folder\\file.py

            for filename in filenames:
                current_file_filepath = f"{root_dirpath}/{filename}"
                current_file_relative_filepath = f"{relative_root_dirpath}/{filename}"

                s3_object_key_name = f"{s3_target_folderpath}{current_file_relative_filepath}"
                self.upload_file_to_s3(filepath=current_file_filepath, object_key_name=s3_object_key_name,
                                  bucket_name=bucket_name, region_name=bucket_region_name)
                print(f"- {current_file_filepath}  - S3 uploaded to path {s3_object_key_name}")


    def create_package(self, app_folder_path: str, to_zip: bool = False, to_s3: bool = False, stage_name: Optional[str] = None,
                       bucket_name: Optional[str] = None, bucket_region_name: Optional[str] = None) -> str:
        if to_zip is False and to_s3 is False:
            raise Exception(f"To create or upload a package you must set the to_zip and/or the to_s3 argument to True.")

        archive_destination_filepath = os.path.join(Path(app_folder_path).parent, f"{Path(app_folder_path).name}.zip")
        click.echo(f"Making an archive from all the files and folders in {app_folder_path} to {archive_destination_filepath}")

        folders_names_to_excludes = [".aws-sam", ".idea", ".git", "__pycache__", "venv", "dist", "node_modules", "lambda_layer",
                                     "speech_synthesis\\export", "speech_synthesis\\polly\\project_data", "libs"]
        # todo: re-include node_modules for production

        has_found_engine_in_project_files = False
        zip_object = None if to_zip is not True else zipfile.ZipFile(file=archive_destination_filepath, mode="w")
        # has_found_framework_in_project_files = False

        if to_s3 is True:
            print(f"Uploading of files starting...  - bucket_name:{bucket_name}  - bucket_region_name:{bucket_region_name}")
        if to_zip is True:
            print(f"Zipping of files starting...   - archive_destination_filepath:{archive_destination_filepath}")

        # Include the projects files
        for root_dirpath, dirs, filenames in os.walk(app_folder_path, topdown=True):
            # The topdown arg allow use to modify the dirs list in the walk, and so we can easily exclude folders.

            # dirs[:] = [dirpath for dirpath in dirs if Path(dirpath).name not in folders_names_to_excludes]
            # This code would only check if the name of the directory is found in the folders names to exclude

            # Where the code below, will check, if the names or paths of all the folder names or paths to exclude,
            # have been found in the dirpath. This approach allow to check for both folder name, and folder paths.
            for folder_name in folders_names_to_excludes:
                for dirname in dirs:
                    dirpath = os.path.join(root_dirpath, dirname)
                    if folder_name in dirpath:
                        dirs.remove(dirname)

            # if Path(root_dirpath).name == "inoft_vocal_engine":
            #    has_found_engine_in_project_files = True

            relative_root_dirpath = root_dirpath.replace(app_folder_path, "").replace("\\", "/").lstrip("/")
            # We need to remove potential leading backslashes and slashes, otherwise, when we will use the os.path.join
            # function, it will remove the first element of the path we are trying to combine the relative_root_dirpath.
            # For example, if we did os.path.join("prod", "\\folder\\file.py") we would end up with \\folder\\file.py
            # where as doing os.path.join("prod", "folder\\file.py") will give us prod\\folder\\file.py
            relative_root_dirpath = f"inoft_vocal_engine/{relative_root_dirpath}"

            for filename in filenames:
                current_file_filepath = f"{root_dirpath}/{filename}"
                current_file_relative_filepath = f"{relative_root_dirpath}/{filename}"
                current_file_print_log = f"- {current_file_filepath}"

                if to_zip is True:
                    zip_object.write(filename=current_file_filepath,
                                     arcname=current_file_relative_filepath)
                    current_file_print_log += f"  - ZIP added to path {current_file_relative_filepath}"

                if to_s3 is True:
                    s3_object_key_name = f"{stage_name}/{current_file_relative_filepath}"
                    self.upload_file_to_s3(filepath=current_file_filepath, object_key_name=s3_object_key_name,
                                      bucket_name=bucket_name, region_name=bucket_region_name)
                    current_file_print_log += f"  - S3 uploaded to path {s3_object_key_name}"

                print(current_file_print_log)

            # todo: when doing a redeploy, check that the lambda layer used, is the right layer for the current framework version

            """
            if has_found_engine_in_project_files is False:
                # On the condition that we have not found the engine in the project files. This function exist both for development sake,
                # where i can deploy with a dev version of the engine without having to publish to pip, while stile using the command line
                # interface from the pip package. And also since the engine (and its correct version) will be included in the deployment
                # package, so that if an user download an app deployed in a different version that the current version of the framework he
                # is using, the version of the framework that will be used will be the one included in its package, not the one installed.

                # We will include the inoft_vocal_engine himself (he is not included in the lambda layers)
                # I do not include the framework in the lambda layer, because it would be weird for someone to update the framework,
                # do a redeploy, and not have upgraded its deploy. Where as we can check if its layer is the right layer for the version
                # of his framework. And mostly because it would be annoying for me to recreate a new lambda layer on each update of the
                # framework, and that he is light enough to be included in the package and the increase in upload time to not be noticeable ;)
                inoft_vocal_engine_folder_path = os.path.dirname(os.path.realpath(inoft_vocal_engine.__file__))
                for root_dirpath, dirs, filenames in os.walk(inoft_vocal_engine_folder_path):
                    relative_dirpath_with_root_included = root_dirpath.replace(str(Path(inoft_vocal_engine_folder_path).parent), "")
                    # We only replace the parent of the framework folder path, because we want it to be in its inoft_vocal_engine folder.
                    for filename in filenames:
                        zip_object.write(filename=os.path.join(root_dirpath, filename),
                                         arcname=os.path.join(relative_dirpath_with_root_included, filename))
            """


        # from shutil import make_archive
        # zip_filepath = make_archive(base_name=archive_destination_filepath_without_extension, format="zip", root_dir=app_folder_path)
        click.echo(f"Archive completed at {click.style(text=archive_destination_filepath, bold=True)}")

        # Warn if this is too large for Lambda.
        file_stats = os.stat(archive_destination_filepath)
        if file_stats.st_size > 52428800:
            click.echo("Warning: Application zip package is likely to be too large for AWS Lambda. Try to make it smaller")

        return archive_destination_filepath
